"""
Image processing functions.
"""
import numpy as np
from PIL import Image,ImageDraw,ImageFont
from .config import *
from pygments import lex
from pygments.lexers import *
def gettext(s, font):
    """
    Get size of text.
    :param s: text
    :param font: font
    :return: size (tuple)
    """
    img = Image.new('RGBA', (1, 1), (0, 0, 0, 0))
    idd = ImageDraw.ImageDraw(img)
    return idd.textsize(s, font)
def create_alpha(token_list,font):
    """
    Create a transparent background image by the token list.
    :param token_list: Token list.
    :param font: font
    :return: a transparent image (as numpy array).
    """
    final_w,final_h=0,0
    for i,j in token_list:
        final_w+=gettext(j,font)[0]
        final_h=max([final_h,gettext(j,font)[1]])
    size=(final_w,final_h)
    return np.array(Image.new('RGBA',size,(0,0,0,0)))
def add_text(alpha,token_list,font,config,token_checker):
    """
    Add tokens to background.
    :param alpha: background image.
    :param token_list: Token List.
    :param font: font.
    :param config: Highlighter config instance.
    :param token_checker: A TokenChecker instance.
    :return: Background with text added.
    """
    alpha=Image.fromarray(alpha)
    pos=0
    idd=ImageDraw.ImageDraw(alpha)
    for i,j in token_list:
        idd.text((pos,0),j,fill=config(token_checker(i)),font=font)
        sw,sh=gettext(j,font)
        pos+=sw
    return np.array(alpha)
def create_lineno(lineno,alpha,font,fgcolor,width,offset):
    """
    Create a image for line number.
    :param lineno: Line number.
    :param alpha: Transparent image
    :param font: font
    :param fgcolor: foreground color
    :param width: image width
    :param offset: Offset
    :return: Line number image.
    """
    a=alpha
    alpha=Image.fromarray(alpha)
    height=a.shape[0]
    bg=Image.new('RGBA',(width,height),(0,0,0,0))
    w,h=gettext(str(lineno),font)
    x=width-w-offset
    y=int(height-h)/2
    idd=ImageDraw.ImageDraw(bg)
    idd.text((x,y),str(lineno),font=font,fill=fgcolor)
    return np.array(bg)
def image_merge_row(img1,img2):
    img3=np.zeros((img1.shape[0],img2.shape[1]+img1.shape[1],4),dtype=np.uint8)
    img3[:,:img1.shape[1],:]=img1
    img3[:,img1.shape[1]:,:]=img2
    return img3
def image_merge_col(img1,img2):
    img3 = np.zeros((img1.shape[0]+img2.shape[0], img1.shape[1], 4), dtype=np.uint8)
    img3[:img1.shape[0],:,  :] = img1
    img3[img1.shape[0]:,:,  :] = img2
    return img3
def image_fit(img,w):
    img2=np.array(Image.new('RGBA',(w,img.shape[0]),(0,0,0,0)))
    img2[:,:img.shape[1],:]=img
    return img2



def split_line(token_list):
    result=[]
    temp=[]
    for i,j in token_list:
        if '\n' not in j:
            temp.append((i,j))
        else:
            if j=='\n':
                result.append(temp)
                temp=[]
                continue
            jsplit=j.split('\n')
            for k in jsplit:
                temp.append((i,k))
                result.append(temp)
                temp=[]
    return result



from PIL import Image,ImageFont
import numpy as np
import cv2
from .imageproc import *
from pygments.lexers import *
from pygments import *
from .config import *
from .easy import *
class Background():
    """
    Class for background of code screenshot.
    """
    def __init__(self,obj):
        if isinstance(obj,tuple): # Background is a color.
            if isinstance(obj[0],tuple):
                self.type=0
                self.lineno_bg=obj[0]
                self.code_bg=obj[1]
            else:
                self.lineno_bg=obj
                self.code_bg=obj
                self.type=0
        else: # Background is an image.
            self.type=1
            self.bg=obj
    def color(self,code,lineno): # Used more numpy functions now, so much faster.
        if self.type:
            whole=image_merge_row(lineno,code)
            bg=cv2.resize(self.bg,(whole.shape[1],whole.shape[0]),interpolation=cv2.INTER_LINEAR)
            mask=(whole[:,:,3]==0)
            whole[:,:,3]-=whole[:,:,3]*mask
            whole[:,:,3]+=255
            whole[:,:,0]-=whole[:,:,0]*mask
            whole[:,:,1]-=whole[:,:,1]*mask
            whole[:,:,2]-=whole[:,:,2]*mask
            whole[:, :, 0] += bg[:, :, 0] * mask
            whole[:, :, 1] += bg[:, :, 1] * mask
            whole[:, :, 2] += bg[:, :, 2] * mask
            return np.array(Image.fromarray(whole).convert('RGB'))
        else:
            mask_code=(code[:,:,3]==0)
            mask_lineno=(lineno[:,:,3]==0)
            code[:,:,3]-=code[:,:,3]*mask_code
            lineno[:,:,3]-=lineno[:,:,3]*mask_lineno
            code[:,:,3]+=255
            lineno[:, :, 3] += 255
            code[:,:,0]-=code[:,:,0]*mask_code
            code[:, :, 0] += np.uint8(self.code_bg[0]) * mask_code
            code[:, :, 1] -= code[:, :, 1] * mask_code
            code[:, :, 1] += np.uint8(self.code_bg[1]) * mask_code
            code[:, :, 2] -= code[:, :, 2] * mask_code
            code[:, :, 2] += np.uint8(self.code_bg[2]) * mask_code
            lineno[:, :, 0] -= lineno[:, :, 0] * mask_lineno
            lineno[:, :, 0] += np.uint8(self.lineno_bg[0]) * mask_lineno
            lineno[:, :, 1] -= lineno[:, :, 1] * mask_lineno
            lineno[:, :, 1] += np.uint8(self.lineno_bg[1]) * mask_lineno
            lineno[:, :, 2] -= lineno[:, :, 2] * mask_lineno
            lineno[:, :, 2] += np.uint8(self.lineno_bg[2]) * mask_lineno
            return np.array(Image.fromarray(image_merge_row(lineno,code)).convert('RGB'))
def fromstr(string):
    def hex2rgb(x):
        return (int(x[:2],16),int(x[2:4],16),int(x[4:6],16))
    if '.' in string:
        return Background(image_open(string))
    elif len(string)==6:
        return Background(hex2rgb(string))
    else:
        return Background((hex2rgb(string[:6]),hex2rgb(string[6:])))



"""
Easy functions for images
"""
from PIL import Image,ImageTk
import numpy as np
def require_tk():
    return __import__('tkinter')
def image_open(path):
    return np.array(Image.open(path))
def image_save(path,img):
    Image.fromarray(img).save(path)
def image_show(img,title=''):
    tk=require_tk()
    image_pil=Image.fromarray(img)
    wn=tk.Tk()
    wn.title(title)
    image_tk = ImageTk.PhotoImage(image_pil)
    wn.geometry('{}x{}'.format(img.shape[1],img.shape[0]))
    l=tk.Label(wn,text='',image=image_tk)
    l.place(x=0,y=0,width=img.shape[1],height=img.shape[0])
    wn.mainloop()



"""
Takes pictures of code!
"""
from pygments import *
from pygments.lexers import *
from .background import *
from PIL import ImageFont
from .config import *
from .imageproc import *
from .easy import *
from .tokenproc import split_line
from argparse import *
from .debug import *
from colorama import *
import sys
__version__='1.0.0'
default_config=ColorConfig(
    cfdct={
        'keyword':(127,127,0),
        'int':(127,0,127),
        'oct':(127,0,127),
        'bin':(127,0,127),
        'hex':(127,0,127),
        'float':(127,0,127),
        'comment':(127,127,127),
        'docstring':(127,127,127),
        'builtin':(0,0,127),
        'string':(127,0,0),
        'shebang':(127,127,127),
        'escape':(255,193,193),
        'char':(139,101,139),
        'affix':(127,0,0),
        'data':(127,0,127),
        'exception':(0,127,127),
        'preprocessor':(0,0,127),
        'preprocessor-file':(40,113,62)
    }
)
def shot(code,config=default_config,lang='auto',background=Background(((255,255,255),(255,255,255))),font=ImageFont.truetype('simsun.ttc',size=30),lineno_fgcolor=(0,0,0),tabsize=4,verbose=False):
    """
    Creates an image of the given code.
    :param code: Code as a string.
    :param config: Syntax highlight config.
    :param lang: Language of the code.
    :param background: Background.
    :param font: Font.
    :param lineno_fgcolor: Line number foreground color.
    :param tabsize: Tab size
    :param verbose: Verbose
    :return: Image as a numpy array.
    """
    if verbose:
        init(autoreset=True)
    if not code.strip():
        debug_output('Source code should NOT be empty',ERROR)
        return
    code=code.replace('\t',' '*tabsize)
    if lang=='auto':
        lang_lex=guess_lexer(code)
    else:
        lang_lex=get_lexer_by_name(lang)
    if verbose:
        debug_output('Tokenizing...',INFO)
    token_list=lex(code,lang_lex)
    token_list_line=split_line(token_list)
    linenos=[]
    lines=[]
    if verbose:
        debug_output('Processing {} lines of tokens'.format(len(token_list_line)),INFO)
    for i,tok in enumerate(token_list_line):
        if verbose:
            debug_output('Creating transparent background for line {}'.format(i+1),INFO)
        alpha=create_alpha(tok,font)
        if verbose:
            debug_output('Adding text to background for line {}'.format(i+1),INFO)
        text=add_text(alpha,tok,font,config,TokenChecker())
        if verbose:
            debug_output('Creating line number {}'.format(i+1),INFO)
        lineno=create_lineno(i+1,alpha,font,lineno_fgcolor,font.size*2,font.size//6)
        lines.append(text)
        linenos.append(lineno)
    line=None
    if verbose:
        debug_output('Start merging code lines',INFO)
    for i,j in enumerate(lines):
        if verbose:
            debug_output('Merging code line {}'.format(i+1),INFO)
        if i==0:
            line=j
        else:
            if line.shape[1]>j.shape[1]:
                j=image_fit(j,line.shape[1])
            else:
                line = image_fit(line, j.shape[1])
            line=image_merge_col(line,j)
    lineno = None
    if verbose:
        debug_output('Start merging line numbers',INFO)
    for i, j in enumerate(linenos):
        if verbose:
            debug_output('Merging line number {}'.format(i+1),INFO)
        if i == 0:
            lineno = j
        else:
            lineno = image_merge_col(lineno, j)
    if verbose:
        debug_output('Adding background',INFO)
    result=background.color(line,lineno)
    if verbose:
        debug_output('Done!',INFO)
    return result


def _cli():
    a=ArgumentParser()
    a.add_argument('-i','--code',help='Source code file name')
    a.add_argument('-V','--version',help='Print the version of sourceshot and exit',action='store_true')
    a.add_argument('-v','--verbose',help='Print out verbose logs',action='store_true')
    a.add_argument('-o','--output',help='Output image file name, the output image will be shown in a tkinter window if not given')
    a.add_argument('-f','--font',help='Font file name',default='simsun.ttc')
    a.add_argument('-s','--fontsize',help='Font size',default='30')
    a.add_argument('-l','--language',help='Source code language, if not given, the language will be automatically determined',default='auto')
    a.add_argument('-c','--config',help='Syntax highlight config JSON file')
    a.add_argument('-b','--background',help='Background color or image, could be:\n1. 6 hexadecimal bits representing the background color of the whole code.\n2. 12 hexadecimal bits, th first 6 represents the background color of the code, the lest 6 represents the background color of the line number.\n3. Path of an image.',default='FFFFFF')
    a.add_argument('-e','--encoding',help='Source code encoding',default='utf-8')
    a.add_argument('-t','--tabsize',help='Tab size',default='4')
    args=a.parse_args()
    if args.version:
        print(__version__)
        return
    if args.verbose:
        verbose=True
    else:
        verbose=False
    if not args.fontsize.isdigit():
        debug_output('Font size should be an integer',ERROR)
        return -1
    try:
        font=ImageFont.truetype(args.font,size=int(args.fontsize))
    except BaseException as e:
        debug_output('Error occurred when creating font instance: {}'.format(e),ERROR)
        return -1
    if args.config:
        config=color_config_from_json(args.config)
    else:
        config=default_config
    
    if args.code:
        try:
            with open(args.code, 'r', encoding=args.encoding) as f:
                source = f.read()
        except UnicodeError:
            debug_output('Found chars which cannot be decoded in the current encoding \'{}\', these characters will be ignored'.format(args.encoding),WARNING)
            with open(args.code, 'r', encoding=args.encoding,errors='ignore') as f:
                source = f.read()
    else:
        debug_output('Read code from stdin',INFO)
        source=sys.stdin.read()
    try:
        tabsize=int(args.tabsize)
    except:
        debug_output('Tab size should be an integer', ERROR)
        return -1
    img=shot(source,config,verbose=verbose,lang=args.language,font=font,tabsize=tabsize,background=fromstr(args.background))
    if not args.output:
        try:
            image_show(img)
        except BaseException as e:
            debug_output('Error occurred when showing image: {}'.format(e), ERROR)
            return -1
    else:
        try:
            image_save(args.output, img)
        except BaseException as e:
            debug_output('Error occurred when saving file: {}'.format(e), ERROR)
            return -1
        debug_output('Successfully saved file as {}'.format(args.output),INFO)

__all__=['shot','default_config']
if __name__=="__main__":
    sys.exit(_cli())


from setuptools import setup
def readme():
    with open('README.md','r') as f:
        return f.read()
setup(
    name='sourceshot',
    version='1.0.1',
    description='Takes pictures of code!',
    long_description=readme(),
    packages=['sshot'],
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts':[
            'sshot=sshot:_cli'
        ]
    },
    python_requires='>=3.6',
    install_requires=['pygments','pillow<10','numpy','opencv-python','colorama'],
    keywords=['code','picture']
)
